#include "common.h"

#define BLOCK_SIZE_X 16
#define BLOCK_SIZE_Y 16
#define COARSEN_FACTOR 2
#define HALO_SIZE_X (BLOCK_SIZE_X+2)
#define HALO_SIZE_Y (BLOCK_SIZE_Y+2)
#define HALF_HALO_SIZE_Y (BLOCK_SIZE_Y/2+1)
#define HALF_TOTAL_HALO_SIZE (HALO_SIZE_X*HALO_SIZE_Y/2)



__global__ void block2D_naive(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	int i = blockIdx.x*blockDim.x+threadIdx.x;
    int j = blockIdx.y*blockDim.y+threadIdx.y;
	if( i>0 && j>0 &&(i<nx-1) &&(j<ny-1) )
	{
		for(int k=1;k<nz-1;k++)
		{
		__syncthreads();
		Anext[Index3D (nx, ny, i, j, k)] = 
		A0[Index3D (nx, ny, i, j, k + 1)] +
		A0[Index3D (nx, ny, i, j, k - 1)] +
		A0[Index3D (nx, ny, i, j + 1, k)] +
		A0[Index3D (nx, ny, i, j - 1, k)] +
		A0[Index3D (nx, ny, i + 1, j, k)] +
		A0[Index3D (nx, ny, i - 1, j, k)]
		- 6.0 * A0[Index3D (nx, ny, i, j, k)] / (fac*fac);
		}
	}
}

__global__ void block2D_opt_1(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	//For optimization 1
}



__global__ void block2D_opt_2(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	//For optimization 2
}

__global__ void block2D_opt_3(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	//For optimization 3
}

__global__ void block2D_shared(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	
	int inx=threadIdx.x;
	int iny=threadIdx.y;
	int i = blockIdx.x*blockDim.x+inx;
	int j = blockIdx.y*blockDim.y+iny;
	
	//linearize thread id
	int id=iny*BLOCK_SIZE_X+inx;
	
	//reshape to shared memory thread ids
	int snx=id%HALO_SIZE_X;
	int sny=id/HALO_SIZE_X;
	int si=blockIdx.x*blockDim.x+snx-1;
	int sj=blockIdx.y*blockDim.y+sny-1;
	
	//malloc shared memory
	__shared__ float sh_A0[HALO_SIZE_Y][HALO_SIZE_X];
	
	sh_A0[iny][inx]=0.0f;
	__syncthreads();
	
	//get active thread ids
	bool thread_active = id<HALF_TOTAL_HALO_SIZE;
	
	//calculate reasonable region for load and store
	bool w_region =  i>0 && j>0 &&(i<nx-1) &&(j<ny-1) ;
	bool r_region =  si>=0 && sj>=0  &&(si<nx) &&(sj<ny) ;
	bool r_region_2 =  si>=0 &&(si<nx) &&((sj+HALF_HALO_SIZE_Y)<ny) ;
	
	

	for(int k=1;k<nz-1;k++)
	{
	
		//load twice to collect whole data
		//because thread number (16*16) is smaller than data size(18*18)
		if(thread_active)
		{
			if(r_region)
			{
				sh_A0[sny][snx]=A0[Index3D (nx, ny, si, sj, k)];
			}
			if(r_region_2)
			{
				sh_A0[sny+HALF_HALO_SIZE_Y][snx]=A0[Index3D (nx, ny, si, sj+HALF_HALO_SIZE_Y, k)];
			}
		}
		__syncthreads();

		if(w_region)
		{
			Anext[Index3D (nx, ny, i, j, k)] =	
			A0[Index3D (nx, ny, i, j, k + 1)] +
			A0[Index3D (nx, ny, i, j, k - 1)] +
			sh_A0[iny+2][inx+1]  +
			sh_A0[iny][inx+1]  +
			sh_A0[iny+1][inx+2]  +
			sh_A0[iny+1][inx] 
			- 6.0f * sh_A0[iny+1][inx+1] / (fac*fac);
		}
		__syncthreads();
	}
}
